# FIXME

import sys
import os
import inspect
import collections
import functools
import re
import pprint
import string
import collections.abc
from enum import Enum, EnumMeta


class DotDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    __reduce__ = lambda self: (dict, (), None, None, iter(self.items()))


class DotDictRef:
    """
    Dot-accessible reference to a dictionary data
    """

    def __new__(cls, data):
        def _getattr(self, key):
            if key in data:
                return data[key]
            else:
                raise AttributeError("key {} does not exist".format(key))

        cls.__getattr__ = _getattr
        cls.__setattr__ = data.__setitem__
        cls.__setattr__ = data.__setitem__
        cls.__delattr__ = data.__delitem__
        cls.__getitem__ = data.__getitem__
        cls.__setitem__ = data.__setitem__
        cls.__delitem__ = data.__delitem__
        cls.__contains__ = data.__contains__
        # make pickleable
        cls.__reduce__ = lambda self: (dict, (), None, None, iter(data.items()))
        cls.__repr__ = data.__repr__
        return super().__new__(cls)


def show_set_diff(data1, data2, name1=None, name2=None):
    if not name1:
        name1 = "set#1"
    if not name2:
        name2 = "set#2"
    data1, data2 = set(data1), set(data2)
    if data1 - data2:
        print(f"Keys in {name1} but not in {name2}: ", list(data1 - data2))
    if data2 - data1:
        print(f"Keys in {name2} but not in {name1}: ", list(data2 - data1))


def _get_qualified_type_name(type_):
    name = str(type_)
    r = re.compile("<class '(.*)'>")
    match = r.match(name)
    if match:
        return match.group(1)
    else:
        return name


def assert_type(x, expected_type, message=""):
    assert isinstance(x, expected_type), (
        message + ": " if message else ""
    ) + "expected type `{}`, actual type `{}`".format(
        _get_qualified_type_name(expected_type), _get_qualified_type_name(type(x))
    )
    return True


def _pformat(obj, **pformat_kwargs):
    "only pformat collections.abc.Container"
    if isinstance(obj, collections.abc.Container) and not isinstance(obj, str):
        indent = pformat_kwargs.pop("indent", 4)  # default indent 4
        return pprint.pformat(obj, indent=indent, **pformat_kwargs)
    else:
        return obj


def prettyprint(*objs, **pformat_kwargs):
    "apply pprint.pformat to only collections.abc.Container"
    objs = [_pformat(o) for o in objs]
    print(*objs, **pformat_kwargs)


def print_debug(*objs, h="", **pformat_kwargs):
    """
    Args:
      *objs: objects to be pretty-printed
      h: header string
      **kwargs: other kwargs to pass on to ``pprint()``
    """
    if h:
        print("=" * 20, h, "=" * 20)
    prettyprint(*objs, **pformat_kwargs)
    if h:
        print("=" * (42 + len(h)))


def print_multiprocess(
    *objs,
    all_prefix="=>",
    master_prefix=None,
    worker_prefix="<{}>",
    process_rank,
    broadcast,
    pretty_format=True,
    print_method=None,
    **print_kwargs,
):
    """
    Used in distributed WorkerManager and Session API

    Args:
        all_prefix (default '=>')
        master_prefix (default None): only active with broadcast=True,
        worker_prefix (default '<{}>'): worker_prefix.format(rank)
        pretty_format (default True): use pprint.pformat
        broadcast (default False): all worker processes print to stdout
    """
    is_master = process_rank == 0
    if not is_master and not broadcast:
        return  # print nothing
    if print_method is None:
        print_method = print

    prefix = all_prefix
    space = " " if all_prefix.strip() else ""
    if master_prefix is None:
        master_prefix = worker_prefix.format(0)
    if is_master:
        if master_prefix and broadcast:
            prefix += space + master_prefix
    else:  # worker
        if not worker_prefix:
            pass
        elif isinstance(worker_prefix, str):
            prefix += space + worker_prefix.format(process_rank)
        else:
            raise ValueError(
                "worker_prefix must be a formatting str, worker_prefix.format(rank: int)"
            )
    if pretty_format:
        objs = [_pformat(o) for o in objs]

    return print_method(prefix, *objs, **print_kwargs)


def printfmt_multiprocess(
    fmt_str,
    *fmt_args,
    all_prefix="=>",
    master_prefix=None,
    worker_prefix="<{}>",
    process_rank,
    broadcast,
    pretty_format=True,
    print_method=None,
    **fmt_kwargs,
):

    if pretty_format:
        fmt_args = [_pformat(o) for o in fmt_args]
        fmt_kwargs = {name: _pformat(value) for name, value in fmt_kwargs.items()}
    s = fmt_str.format(*fmt_args, **fmt_kwargs)
    return print_multiprocess(
        s,
        all_prefix=all_prefix,
        master_prefix=master_prefix,
        worker_prefix=worker_prefix,
        process_rank=process_rank,
        broadcast=broadcast,
        pretty_format=False,
        print_method=print_method,
    )


def get_print_multiprocess_options(option_dict):
    if not option_dict:
        option_dict = {}
    defaults = {
        "all_prefix": "=>",
        "master_prefix": None,
        "worker_prefix": "<{}>",
        "pretty_format": True,
        "broadcast": False,
    }
    defaults.update(option_dict)  # overwrite any defaults with values in option
    return defaults


class _GetItemEnumMeta(EnumMeta):
    """
    Hijack the __getitem__ method from metaclass, because subclass cannot
        override magic methods. More informative error message.
    """

    def __getitem__(self, option):
        enum_class = None
        for v in self.__members__.values():
            enum_class = v.__class__
            break
        assert enum_class is not None, "must have at least one option in StringEnum"
        return get_enum(enum_class, option)


class StringEnum(Enum, metaclass=_GetItemEnumMeta):
    """
    https://docs.python.org/3.4/library/enum.html#duplicatefreeenum
    The created options will automatically have the same string value as name.

    Support [] subscript, i.e. MyFruit['orange'] -> MyFruit.orange
    """

    def __init__(self, *args, **kwargs):
        self._value_ = self.name


def create_string_enum(class_name, option_names):
    assert_type(option_names, str)
    assert_type(option_names, list)
    return StringEnum(class_name, option_names)


def get_enum(enum_class, option):
    """
    Args:
        enum_class:
        option: if the value doesn't belong to Enum, throw error.
            Can be either the str name or the actual enum value
    """
    assert issubclass(enum_class, StringEnum)
    if isinstance(option, enum_class):
        return option
    else:
        assert_type(option, str)
        option = option.lower()
        options = enum_class.__members__
        if option not in options:
            raise ValueError(
                '"{}" is not a valid option for {}. '
                "Available options are {}.".format(
                    option, enum_class.__name__, list(options)
                )
            )
        return options[option]


def fformat(float_num, precision):
    """
    https://stackoverflow.com/a/44702621/3453033
    """
    assert isinstance(precision, int) and precision > 0
    return "{{:.{}f}}".format(precision).format(float_num).rstrip("0").rstrip(".")


def make_tuple(elem, repeats):
    """
    E.g. expand 3 into (3, 3) for things like strides/paddings
    """
    if isinstance(elem, int):
        return [elem] * repeats
    else:
        assert len(elem) == repeats
        return [int(x) for x in elem]


def iter_last(iterable):
    """
    For processing the last element differently
    Yields: (is_last=bool, element)
    """
    length = len(iterable)
    return ((i == length - 1, x) for i, x in enumerate(iterable))


def merge_dicts(*dicts, descending_priority=True):
    """
    Merge all dicts into a single new dict.
    Args:
        descending_priority: deals with duplicate key.
            from left (earlier) to right (later) dict args,
            if True, the earlier dicts' duplicate key will override later dicts.
    """
    for d in dicts:
        assert isinstance(d, dict)
    if descending_priority:
        dicts = reversed(dicts)
    D = {}
    for d in dicts:
        D.update(d)
    return D


def parse_formatter_vars(format_str):
    """
    Returns:
        list of all variables that appear in the format string
        e.g. '{foo:.2f} has {bar:>5s}' returns ['foo'], ['bar']
    """
    names = []
    for _, name, fmt, _ in string.Formatter().parse(format_str):
        # the last fmt is always None with no corresponding name
        if fmt is not None:
            names.append(name)
    return names


def case_insensitive_match(items, key):
    """
    Args:
        items: iterable of keys
        key: search for the key case-insensitively

    Returns:
        matched original key (with cases), None if no match
    """
    for k in items:
        if k.lower() == key.lower():
            return k
    return None


def _to_float(x):
    # make an attempt to convert to plain float. If fails, just keep original
    if isinstance(x, int):
        return x
    try:
        return float(x)
    except (ValueError, TypeError):
        return x


def to_float_dict(d):
    "recursive"
    newd = {}
    for k, v in d.items():
        if isinstance(v, dict):
            newd[k] = to_float_dict(v)
        else:
            newd[k] = _to_float(v)
    return newd


def _get_bound_args(func, *args, **kwargs):
    """
    https://docs.python.org/3/library/inspect.html#inspect.BoundArguments
    def f(a, b, c=5, d=6): pass
    get_bound_args(f, 3, 6, d=100) -> {'a':3, 'b':6, 'c':5, 'd':100}

    Returns:
        OrderedDict of bound arguments
    """
    arginfo = inspect.signature(func).bind(*args, **kwargs)
    arginfo.apply_defaults()
    return arginfo.arguments


class _Deprecated_SaveInitArgsMeta(type):
    """
    Bounded arguments:
    https://docs.python.org/3/library/inspect.html#inspect.BoundArguments

    Store the captured constructor arguments to <instance>._init_args_dict
    as OrderedDict. Can be retrieved by the property method <obj>.init_args_dict
    Includes both the positional args (with the arg name) and kwargs
    """

    def __init__(cls, name, bases, attrs):
        # WARNING: must add class method AFTER super.__init__
        # adding attrs['new-method'] before __init__ has no effect!
        super().__init__(name, bases, attrs)

        @property
        def init_args_dict(self):
            return self._init_args_dict

        cls.init_args_dict = init_args_dict

    def __call__(cls, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        try:
            obj._init_args_dict = _get_bound_args(obj.__init__, *args, **kwargs)
        except TypeError:  # __init__ has special stuff like *args
            obj._init_args_dict = None
        return obj


class SaveInitArgsMeta(type):
    """
    Store __init__ call args in self.init_args_dict
    Positional args to __init__ will be stored under `None` key
    Keyword args to __init__ will be stored as-is in init_args_dict
    """

    def __init__(cls, name, bases, attrs):
        # WARNING: must add class method AFTER super.__init__
        # adding attrs['new-method'] before __init__ has no effect!
        super().__init__(name, bases, attrs)

        @property
        def init_args_dict(self):
            return self._init_args_dict

        cls.init_args_dict = init_args_dict

    def __call__(cls, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        init_dict = {}
        if args:
            init_dict[None] = list(args)  # convert to list for YAML compatibility
        init_dict.update(kwargs)
        obj._init_args_dict = init_dict
        return obj


class SaveInitArgs(metaclass=SaveInitArgsMeta):
    """
    Either use metaclass hook:
        class MyObj(metaclass=SaveInitArgsMeta)
    or simply inherit
        class MyObj(SaveInitArgs)
    """

    pass


class AutoInitializeMeta(type):
    """
    Call the special method ._initialize() after __init__.
    Useful if some logic must be run after the object is constructed.
    For example, the following code doesn't work because `self.y` does not exist
    when super class calls self._initialize()

    class BaseClass():
        def __init__(self):
            self._initialize()

        def _initialize():
            self.x = self.get_x()

        def get_x(self):
            # abstract method that only subclass

    class SubClass(BaseClass):
        def __init__(self, y):
            super().__init__()
            self.y = y

        def get_x(self):
            return self.y * 3

    Fix:
    class BaseClass(metaclass=AutoInitializeMeta):
        def __init__(self):
            pass
            # self._initialize() is now automatically called after __init__

        def _initialize():
            print('INIT', self.x)

        def get_x(self):
            # abstract method that only subclass
            raise NotImplementedError
    """

    def __call__(self, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        assert hasattr(
            obj, "_initialize"
        ), "AutoInitializeMeta requires that subclass implements _initialize()"
        obj._initialize()
        return obj


class SerializedDict:
    def __init__(self, keys):
        self.keys = keys

    def dump(self):
        return {key: getattr(self, key) for key in self.keys}

    def load(self, state_dict):
        for key in self.keys:
            assert key in state_dict, 'Loading error: key "{}" missing'.format(key)
            setattr(self, key, state_dict[key])
        return self


class noop_context:
    """
    Placeholder context manager that does nothing.
    We could have written simply as:

    @contextmanager
    def noop_context(*args, **kwargs):
        yield

    but the returned context manager cannot be called twice, i.e.
    my_noop = noop_context()
    with my_noop:
        do1()
    with my_noop: # trigger generator error
        do2()
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def meta_wrap(decor):
    """
    a decorator decorator, allowing the wrapped decorator to be used as:
    @decorator(*args, **kwargs)
    def callable()
      -- or --
    @decorator  # without parenthesis, args and kwargs will use default
    def callable()

    Args:
      decor: a decorator whose first argument is a callable (function or class
        to be decorated), and the rest of the arguments can be omitted as default.
        decor(f, ... the other arguments must have default values)

    Warning:
      decor can NOT be a function that receives a single, callable argument.
      See stackoverflow: http://goo.gl/UEYbDB
    """
    single_callable = (
        lambda args, kwargs: len(args) == 1 and len(kwargs) == 0 and callable(args[0])
    )

    @functools.wraps(decor)
    def new_decor(*args, **kwargs):
        if single_callable(args, kwargs):
            # this is the double-decorated f.
            # It should not run on a single callable.
            return decor(args[0])
        else:
            # decorator arguments
            return lambda real_f: decor(real_f, *args, **kwargs)

    return new_decor


@meta_wrap
def deprecated(func, msg="", action="warning"):
    """
    Function/class decorator: designate deprecation.

    Args:
      msg: string message.
      action: string mode
      - 'warning': (default) prints `msg` to stderr
      - 'noop': do nothing
      - 'raise': raise DeprecatedError(`msg`)
    """
    action = action.lower()
    if action not in ["warning", "noop", "raise"]:
        raise ValueError("unknown action type {}".format(action))
    if not msg:
        msg = "This is a deprecated feature."

    # only does the deprecation when being called
    @functools.wraps(func)
    def _deprecated(*args, **kwargs):
        if action == "warning":
            print(msg, file=sys.stderr)
        elif action == "raise":
            raise DeprecationWarning(msg)
        return func(*args, **kwargs)

    return _deprecated


def pack_varargs(args):
    """
    Pack *args or a single list arg as list

    def f(*args):
        arg_list = pack_varargs(args)
        # arg_list is now packed as a list
    """
    assert isinstance(args, tuple), "please input the tuple `args` as in *args"
    if len(args) == 1 and isinstance(args[0], (list, tuple)):
        return args[0]
    else:
        return args


def enable_list_arg(func):
    """
    Function decorator.
    If a function only accepts varargs (*args),
    make it support a single list arg as well
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        args = pack_varargs(args)
        return func(*args, **kwargs)

    return wrapper


def enable_varargs(func):
    """
    Function decorator.
    If a function only accepts a list arg, make it support varargs as well
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        args = pack_varargs(args)
        return func(args, **kwargs)

    return wrapper


def pack_kwargs(args, kwargs):
    """
    Pack **kwargs or a single dict arg as dict

    def f(*args, **kwargs):
        kwdict = pack_kwargs(args, kwargs)
        # kwdict is now packed as a dict
    """
    if len(args) == 1 and isinstance(args[0], dict):
        assert not kwargs, "cannot have both **kwargs and a dict arg"
        return args[0]  # single-dict
    else:
        assert not args, "cannot have positional args if **kwargs exist"
        return kwargs


def enable_dict_arg(func):
    """
    Function decorator.
    If a function only accepts varargs (*args),
    make it support a single list arg as well
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        kwargs = pack_kwargs(args, kwargs)
        return func(**kwargs)

    return wrapper


def enable_kwargs(func):
    """
    Function decorator.
    If a function only accepts a dict arg, make it support kwargs as well
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        kwargs = pack_kwargs(args, kwargs)
        return func(kwargs)

    return wrapper


def method_decorator(decorator):
    """
    Decorator of decorator: transform a decorator that only works on normal
    functions to a decorator that works on class methods
    From Django form: https://goo.gl/XLjxKK
    """

    @functools.wraps(decorator)
    def wrapped_decorator(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            def bound_func(*args2, **kwargs2):
                return method(self, *args2, **kwargs2)

            return decorator(bound_func)(*args, **kwargs)

        return wrapper

    return wrapped_decorator


def accepts_varargs(func):
    """
    If a function accepts *args
    """
    params = inspect.signature(func).parameters
    return any(
        param.kind == inspect.Parameter.VAR_POSITIONAL for param in params.values()
    )


def accepts_kwargs(func):
    """
    If a function accepts **kwargs
    """
    params = inspect.signature(func).parameters
    return any(param.kind == inspect.Parameter.VAR_KEYWORD for param in params.values())


def is_signature_compatible(func, *args, **kwargs):
    sig = inspect.signature(func)
    try:
        sig.bind(*args, **kwargs)
        return True
    except TypeError:
        return False


class OrderedSet(collections.abc.MutableSet):
    """
    Recommended recipe from official python:
    https://code.activestate.com/recipes/576694/
    """

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]  # sentinel node for doubly linked list
        self.map = {}  # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError("set is empty")
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return "%s()" % (self.__class__.__name__,)
        return "%s(%r)" % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)



import os
import logging
from typing import Union, Optional

try:
    import colorlog

    IS_COLORLOG_INSTALLED = True
except ImportError:
    IS_COLORLOG_INSTALLED = False


__all__ = [
    "omlet_logger",
    "initialize_omlet_logging",
    "set_logging_level",
    "get_logging_level",
    "create_logging_file_handler",
]


# custom debugging level that's higher than usual to distinguish from
# debug logs from other applications
DEBUG2 = logging.DEBUG + 2
logging.DEBUG2 = DEBUG2

# custom info level that's lower than usual to enable more verbosity
INFOV = logging.INFO - 1
INFOVV = logging.INFO - 2
INFOVVV = logging.INFO - 3
logging.INFOV = INFOV
logging.INFOVV = INFOVV
logging.INFOVVV = INFOVVV

# https://stackoverflow.com/questions/2183233/how-to-add-a-custom-loglevel-to-pythons-logging-facility
logging.addLevelName(DEBUG2, "DEBUG2")
logging.addLevelName(INFOV, "INFOV")
logging.addLevelName(INFOVV, "INFOVV")
logging.addLevelName(INFOVVV, "INFOVVV")


_DATEFMT = "%m-%d-%y %H:%M:%S"


def _gen_logging_function(level):
    def _log_func(self, message, *args, **kws):
        if self.isEnabledFor(level):
            # Yes, logger takes its '*args' as 'args'.
            self._log(level, message, args, **kws)

    return _log_func


logging.Logger.debug2 = _gen_logging_function(DEBUG2)
logging.Logger.infov = _gen_logging_function(INFOV)
logging.Logger.infovv = _gen_logging_function(INFOVV)
logging.Logger.infovvv = _gen_logging_function(INFOVVV)


omlet_logger = logging.getLogger("omlet")


def initialize_omlet_logging(
    file_path=None,
    level=None,
    replace_all_stream_handlers: bool = True,
    enable: bool = True,
):
    """
    Directly manipulate root logger

    Args
        file_path: log file path
        replace_all_stream_handlers: replace StreamHandler of all existing loggers
    """
    if not enable:
        return
    color_handler = create_colorlog_handler()
    root = logging.getLogger()
    root.handlers = []
    if color_handler is not None:
        root.handlers.append(color_handler)
    if file_path:
        root.handlers.append(create_logging_file_handler(file_path))
    set_logging_level(level)

    if replace_all_stream_handlers:
        for name, logger in all_loggers().items():
            if not hasattr(logger, "handlers"):
                continue
            for handler in logger.handlers.copy():
                # we still keep file streaming for other apps
                if isinstance(handler, logging.StreamHandler) and not isinstance(
                    handler, logging.FileHandler
                ):
                    # print(name, 'HANDLER STREAM REMOVED')
                    logger.removeHandler(handler)


def create_colorlog_handler():
    if IS_COLORLOG_INSTALLED:
        # https://pypi.org/project/colorlog/
        handler = colorlog.StreamHandler()
        handler.setFormatter(
            colorlog.ColoredFormatter(
                "[%(cyan)s%(asctime)s%(reset)s][%(blue)s%(name)s%(reset)s][%(log_color)s%(levelname)s%(reset)s] %(message)s",
                # available color prefixes: bold_, thin_, bg_, fg_
                # colors: black, red, green, yellow, blue, purple, cyan and white
                log_colors={
                    "DEBUG": "purple",
                    "DEBUG2": "purple,bg_yellow",
                    "INFOV": "thin_green",
                    "INFOVV": "thin_green",
                    "INFOVVV": "thin_green",
                    "INFO": "green",
                    "WARNING": "yellow",
                    "ERROR": "bold_red",
                    "CRITICAL": "red,bg_white",
                },
                datefmt=_DATEFMT,
            )
        )
        return handler
    else:
        return None


def all_loggers():
    return logging.Logger.manager.loggerDict


def create_logging_file_handler(file_path):
    file_path = os.path.expanduser(file_path)
    handler = logging.FileHandler(file_path)
    handler.setFormatter(
        logging.Formatter(
            "[%(asctime)s][%(name)s][%(levelname)s] %(message)s", datefmt=_DATEFMT
        )
    )
    return handler


def get_logging_level(name=None):
    return logging.getLogger(name).getEffectiveLevel()


def set_logging_level(level: Optional[Union[int, str]] = "info", name=None):
    if level is None:
        return
    if isinstance(level, str):
        level = level.upper()
    logging.getLogger(name).setLevel(level)



import os
import sys
import omlet.utils as U
from pytorch_lightning.callbacks import Callback
from . import omlet_logger


class SourceCodeBackup(Callback):
    DEFAULT_INCLUDE_PATTERNS = ("*.py", "*.sh", "*.yaml", "*.yml", "*.md")

    def __init__(self, source_dir=".", sub_dir="code", include_patterns=None):
        """
        Args:
            tag_var_map: maps a tensorboard tag to the tracked variable
        """
        self.exp_dir = None  # will be populated by OmletTrainer
        self.sub_dir = sub_dir
        self.source_dir = self._get_code_path(source_dir)
        assert os.path.exists(
            self.source_dir
        ), 'source code dir "{}" does not exist'.format(self.source_dir)
        assert os.path.isdir(
            self.source_dir
        ), f'source code dir "{self.source_dir}" is not a directory'
        self._source_dir_last_part = U.last_part_in_path(self.source_dir)
        self._backup_dir = None
        if not include_patterns:
            include_patterns = self.DEFAULT_INCLUDE_PATTERNS
        self.include_patterns = include_patterns

    @property
    def backup_dir(self):
        """
        {exp_dir}/code/{your_repo_name}
        """
        if self._backup_dir is not None:
            return self._backup_dir

        assert self.exp_dir, "please use this callback with OmletTrainer"
        desired_path = U.f_join(self.exp_dir, self.sub_dir, self._source_dir_last_part)
        self._backup_dir = U.next_available_file_name(desired_path)
        if self._backup_dir != desired_path:
            omlet_logger.warning(
                f'Source code backup dir "{U.last_part_in_path(desired_path)}/" already exists, saving instead to "{U.last_part_in_path(self._backup_dir)}/"'
            )
        return self._backup_dir

    def _get_code_path(self, path):
        "handles both abspath and relative path"
        path = os.path.expanduser(path)
        if os.path.isabs(path):
            return path
        else:
            # relative to the launch script
            path = os.path.join(U.script_dir(), path)
            return os.path.abspath(path)

    def on_init_start(self, trainer):
        U.f_mkdir(self.backup_dir)
        U.f_copytree(
            self.source_dir,
            self.backup_dir,
            include=self.include_patterns,
            exist_ok=True,
        )
        omlet_logger.info(
            f"Backed up source code {self.source_dir} to {self.backup_dir}"
        )


class CsvRecord(Callback):
    # TODO
    pass



"""
File system utils.
"""
import os
import sys
import errno
import shutil
import glob
import pwd
import codecs
import hashlib
import tarfile
import fnmatch
from typing import Optional, Union, Callable
from datetime import datetime
from socket import gethostname
from .print_utils import format_str


f_ext = os.path.splitext

f_size = os.path.getsize

is_file = os.path.isfile

is_dir = os.path.isdir

get_dir = os.path.dirname


def owner_name(filepath):
    """
    Returns: file owner name, unix only
    """
    return pwd.getpwuid(os.stat(filepath).st_uid).pw_name


def host_name():
    "Get host name, alias with ``socket.gethostname()``"
    return gethostname()


def host_id():
    """
    Returns: first part of hostname up to '.'
    """
    return host_name().split(".")[0]


def utf_open(fname, mode):
    """
    Wrapper for codecs.open
    """
    return codecs.open(fname, mode=mode, encoding="utf-8")


def is_txt(fpath):
    "Test if file path is a text file"
    _, ext = f_ext(fpath)
    return ext == ".txt"


def f_expand(path):
    return os.path.expandvars(os.path.expanduser(path))


def f_exists(path):
    return os.path.exists(f_expand(path))


def f_join(*fpaths):
    """
    join file paths and expand special symbols like `~` for home dir
    """
    return f_expand(os.path.join(*fpaths))


def f_mkdir(*fpaths):
    """
    Recursively creates all the subdirs
    If exist, do nothing.
    """
    os.makedirs(f_join(*fpaths), exist_ok=True)


def f_mkdir_in_path(fpath):
    """
    fpath is a file,
    recursively creates all the parent dirs that lead to the file
    If exist, do nothing.
    """
    os.makedirs(get_dir(f_expand(fpath)), exist_ok=True)


def last_part_in_path(fpath):
    """
    https://stackoverflow.com/questions/3925096/how-to-get-only-the-last-part-of-a-path-in-python
    """
    return os.path.basename(os.path.normpath(f_expand(fpath)))


def is_abs_path(path):
    return os.path.isabs(f_expand(path))


def is_relative_path(path):
    return not is_abs_path(path)


def f_time(fpath):
    "File modification time"
    return str(os.path.getctime(fpath))


def f_append_before_ext(fpath, suffix):
    """
    Append a suffix to file name and retain its extension
    """
    name, ext = f_ext(fpath)
    return name + suffix + ext


def f_add_ext(fpath, ext):
    """
    Append an extension if not already there
    Args:
      ext: will add a preceding `.` if doesn't exist
    """
    if not ext.startswith("."):
        ext = "." + ext
    if fpath.endswith(ext):
        return fpath
    else:
        return fpath + ext


def f_remove(fpath):
    """
    If exist, remove. Supports both dir and file. Supports glob wildcard.
    """
    fpath = f_expand(fpath)
    for f in glob.glob(fpath):
        try:
            shutil.rmtree(f)
        except OSError as e:
            if e.errno == errno.ENOTDIR:
                try:
                    os.remove(f)
                except:  # final resort safeguard
                    pass


def f_copy(fsrc, fdst, exists_ok=False):
    """
    Supports both dir and file. Supports glob wildcard.
    """
    fsrc, fdst = f_expand(fsrc), f_expand(fdst)
    for f in glob.glob(fsrc):
        try:
            f_copytree(f, fdst, exist_ok=exists_ok)
        except OSError as e:
            if e.errno == errno.ENOTDIR:
                shutil.copy(f, fdst)
            else:
                raise


def _f_copytree(
    src,
    dst,
    symlinks=False,
    ignore=None,
    exist_ok=True,
    copy_function=shutil.copy2,
    ignore_dangling_symlinks=False,
):
    """Copied from python standard lib shutil.copytree
    except that we allow exist_ok
    Use f_copytree as entry
    """
    names = os.listdir(src)
    if ignore is not None:
        ignored_names = ignore(src, names)
    else:
        ignored_names = set()

    os.makedirs(dst, exist_ok=exist_ok)
    errors = []
    for name in names:
        if name in ignored_names:
            continue
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if os.path.islink(srcname):
                linkto = os.readlink(srcname)
                if symlinks:
                    # We can't just leave it to `copy_function` because legacy
                    # code with a custom `copy_function` may rely on copytree
                    # doing the right thing.
                    os.symlink(linkto, dstname)
                    shutil.copystat(srcname, dstname, follow_symlinks=not symlinks)
                else:
                    # ignore dangling symlink if the flag is on
                    if not os.path.exists(linkto) and ignore_dangling_symlinks:
                        continue
                    # otherwise let the copy occurs. copy2 will raise an error
                    if os.path.isdir(srcname):
                        _f_copytree(
                            srcname, dstname, symlinks, ignore, exist_ok, copy_function
                        )
                    else:
                        copy_function(srcname, dstname)
            elif os.path.isdir(srcname):
                _f_copytree(srcname, dstname, symlinks, ignore, exist_ok, copy_function)
            else:
                # Will raise a SpecialFileError for unsupported file types
                copy_function(srcname, dstname)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((srcname, dstname, str(why)))
    try:
        shutil.copystat(src, dst)
    except OSError as why:
        # Copying file access times may fail on Windows
        if getattr(why, "winerror", None) is None:
            errors.append((src, dst, str(why)))
    if errors:
        raise shutil.Error(errors)
    return dst


def _include_patterns(*patterns):
    """Factory function that can be used with copytree() ignore parameter.

    Arguments define a sequence of glob-style patterns
    that are used to specify what files to NOT ignore.
    Creates and returns a function that determines this for each directory
    in the file hierarchy rooted at the source directory when used with
    shutil.copytree().
    """

    def _ignore_patterns(path, names):
        keep = set(
            name for pattern in patterns for name in fnmatch.filter(names, pattern)
        )
        ignore = set(
            name
            for name in names
            if name not in keep and not os.path.isdir(os.path.join(path, name))
        )
        return ignore

    return _ignore_patterns


def f_copytree(src, dst, symlinks=False, ignore=None, include=None, exist_ok=False):
    assert (ignore is None) or (
        include is None
    ), "ignore= and include= are mutually exclusive"
    if ignore:
        ignore = shutil.ignore_patterns(*ignore)
    elif include:
        ignore = _include_patterns(*include)
    _f_copytree(src, dst, ignore=ignore, symlinks=symlinks, exist_ok=exist_ok)


def f_move(fsrc, fdst):
    fsrc, fdst = f_expand(fsrc), f_expand(fdst)
    for f in glob.glob(fsrc):
        shutil.move(f, fdst)


def f_split_path(fpath, normpath=True):
    """
    Splits path into a list of its component folders

    Args:
        normpath: call os.path.normpath to remove redundant '/' and
            up-level references like ".."
    """
    if normpath:
        fpath = os.path.normpath(fpath)
    allparts = []
    while 1:
        parts = os.path.split(fpath)
        if parts[0] == fpath:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == fpath:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            fpath = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def script_dir():
    """
    Returns: the dir of current script
    """
    return os.path.dirname(os.path.realpath(sys.argv[0]))


def parent_dir(location, abspath=False):
    """
    Args:
      location: current directory or file

    Returns:
        parent directory absolute or relative path
    """
    _path = os.path.abspath if abspath else os.path.relpath
    return _path(f_join(location, os.pardir))


def md5_checksum(fpath):
    """
    File md5 signature
    """
    hash_md5 = hashlib.md5()
    with open(f_expand(fpath), "rb") as f:
        for chunk in iter(lambda: f.read(65536), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def make_tar(source_file, output_tarball, compress_mode="gz"):
    """
    Args:
        source_file: source file or folder
        output_tarball: output tar file name
        compress_mode: "gz", "bz2", "xz" or "" (empty for uncompressed write)
    """
    source_file, output_tarball = f_expand(source_file), f_expand(output_tarball)
    assert compress_mode in ["gz", "bz2", "xz", ""]
    with tarfile.open(output_tarball, "w:" + compress_mode) as tar:
        tar.add(source_file, arcname=os.path.basename(source_file))


def extract_tar(source_tarball, output_dir=".", members=None):
    """
    Args:
        source_tarball: extract members from archive
        output_dir: default to current working dir
        members: must be a subset of the list returned by getmembers()
    """
    source_tarball, output_dir = f_expand(source_tarball), f_expand(output_dir)
    with tarfile.open(source_tarball, "r:*") as tar:
        tar.extractall(output_dir, members=members)


def move_with_backup(path, suffix=".bak"):
    """
    Ensures that a path is not occupied. If there is a file, rename it by
    adding @suffix. Resursively backs up everything.

    Args:
        path: file path to clear
        suffix: Add to backed up files (default: {'.bak'})
    """
    path = str(path)
    if os.path.exists(path):
        move_with_backup(path + suffix)
        shutil.move(path, path + suffix)


def insert_before_ext(name, insert):
    """
    log.txt -> log.ep50.txt
    """
    name, ext = os.path.splitext(name)
    return name + insert + ext


def timestamp_file_name(fname):
    timestr = datetime.now().strftime("_%H-%M-%S_%m-%d-%y")
    return insert_before_ext(fname, timestr)


def next_available_file_name(
    file_path,
    suffix_template: Union[str, Callable[[int], str]] = "_v{i+1}",
    before_ext: bool = True,
):
    """
    Args:
        suffix_template: a format string using "i" variable or
            lambda int -> str
        before_ext: True to insert suffix before the extension
    """
    orig_file_path = f_expand(file_path)
    i = 0
    file_path = orig_file_path
    while os.path.exists(file_path):
        if isinstance(suffix_template, str):
            suffix = format_str(suffix_template, i=i)
        elif callable(suffix_template):
            suffix = suffix_template(i)
            assert isinstance(suffix, str)
        else:
            raise NotImplementedError(f"Unsupported suffix template {suffix_template}")
        if before_ext:
            file_path = insert_before_ext(orig_file_path, suffix)
        else:
            file_path = orig_file_path + suffix
        i += 1
    return file_path



import torch


def accuracy(output, target, topk=(1,), scale_100=False):
    """
    Computes the accuracy over the k top predictions for the specified values of k.
    Accuracy is a float between 0.0 and 1.0
    """
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        mult = 100.0 if scale_100 else 1.0
        res = []
        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(float(correct_k.mul_(mult / batch_size).item()))
        return res


class AverageMeter:
    """Computes and stores the average and current value"""

    def __init__(self, name=None, formatter=".2f", keep_history=False):
        """
        Args:
            keep_history: if True, keeps a list of [(value, n)]
        """
        self.name = name
        self.fmt = formatter.lstrip(":")
        self.keep_history = keep_history
        self.reset()

    def reset(self):
        self.history = []
        self.sum = 0
        self.size = 0
        self.n = 0

    def update(self, value, size=1):
        value = float(value)
        size = int(size)
        self.sum += value * size
        self.size += size
        self.n += 1
        if self.keep_history:
            self.history.append((value, size))

    @property
    def value(self):
        if self.size == 0:
            return 0.0
        return self.sum / self.size

    def __float__(self):
        return float(self.value)

    def __str__(self):
        # fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        if self.name:
            fmtstr = "{name} {avg:" + self.fmt + "}"
        else:
            fmtstr = "{avg:" + self.fmt + "}"
        return fmtstr.format(name=self.name, avg=self.value)



import logging
import functools
from typing import Union, Dict, List, Optional
import omlet.utils as U

from pytorch_lightning.callbacks import Callback, ProgressBar
from pytorch_lightning.utilities import rank_zero_warn, rank_zero_only
from .core import ExtendedModule, STAGES


class SummaryMessage(Callback):
    def __init__(
        self,
        train_message: str = None,
        val_message: str = None,
        test_message: str = None,
    ):
        self.timers = {}
        for stage in STAGES:
            self.timers[stage] = U.Timer()
        self._summary_messages = {}
        for stage, msg in zip(STAGES, (train_message, val_message, test_message)):
            self._summary_messages[stage] = msg

    def on_epoch_start(self, trainer, pl_module):
        assert isinstance(pl_module, ExtendedModule)
        # on_train_start and on_train_end enclose the entire training, not a single epoch
        self.timers["train"].start()

    def on_validation_start(self, trainer, pl_module):
        self.timers["val"].start()

    def on_test_start(self, trainer, pl_module):
        self.timers["test"].start()

    def _print_summary_message(self, stage, pl: ExtendedModule):
        if not pl._is_training_started or not self._summary_messages[stage]:
            return
        elapsed = self.timers[stage].elapsed_str()
        msg = (
            f"{stage.capitalize()} summary [{pl.current_epoch:>2}] "
            + self._summary_messages[stage]
            + f" ({stage} time {elapsed})"
        )
        msg = msg.format(epoch=pl.current_epoch, **pl._metrics_history[-1][stage])
        pl.log_info(msg)

    def on_epoch_end(self, trainer, pl_module):
        # WARNING: on_train_end() is when the entire trainer.fit() finishes
        # PTL train_epoch_end is always after val_epoch_end
        self._print_summary_message("train", pl_module)

    def on_validation_end(self, trainer, pl_module):
        self._print_summary_message("val", pl_module)

    def on_test_end(self, trainer, pl_module):
        self._print_summary_message("test", pl_module)


class ExtendedProgressBar(ProgressBar):
    """
    Changes compared to the default one:
        - option to remove `v_num`

    """

    def __init__(
        self,
        remove_v_num=True,
        include_metrics: Union[str, List[str], None] = None,
        exclude_metrics: Union[str, List[str], None] = None,
        formatter: Dict[str, str] = None,
        use_short_name=True,
        *args,
        **kwargs,
    ):
        """
        Args:
            use_short_name: True for "val/acc1" -> "v:acc1"
            include_metrics: list of patterns, e.g. val/*
            exclude_metrics: list of patterns, e.g. train/*
                include takes precedence over exclude

        Super class progress bar args:
            refresh_rate: int = 1
            process_position: int = 0
        """
        super().__init__(*args, **kwargs)
        self._remove_v_num = remove_v_num
        self._include_metrics = include_metrics
        self._exclude_metrics = exclude_metrics
        self._use_short_name = use_short_name
        if formatter is None:
            formatter = {}
        self._formatter = formatter

    def _process_progress_dict(self, trainer):
        progress_dict = trainer.progress_bar_dict.copy()

        if self._remove_v_num:
            if "v_num" in progress_dict:
                progress_dict.pop("v_num")
        for k, v in progress_dict.copy().items():
            if k in self._formatter:
                progress_dict[k] = self._formatter[k].format(v)
            # handle exclusions
            if not U.match_patterns(
                k,
                include=self._include_metrics,
                exclude=self._exclude_metrics,
                precedence="exclude",
            ):
                progress_dict.pop(k)
            elif "/" in k and self._use_short_name:
                prefix, mname = k.split("/", 1)
                progress_dict[f"{prefix[0]}:{mname}"] = progress_dict[k]
                progress_dict.pop(k)
        # self.main_progress_bar.set_description(desc)
        return progress_dict

    def on_batch_end(self, trainer, pl_module):
        progress_dict = self._process_progress_dict(trainer)
        # super().on_batch_end(trainer, pl_module)
        # directly copied from pytorch_lightning/callbacks/progress
        if self.is_enabled and self.train_batch_idx % self.refresh_rate == 0:
            self.main_progress_bar.update(self.refresh_rate)
            self.main_progress_bar.set_postfix(**progress_dict)

    def on_validation_end(self, trainer, pl_module):
        progress_dict = self._process_progress_dict(trainer)
        # directly copied from pytorch_lightning/callbacks/progress
        self.main_progress_bar.set_postfix(**progress_dict)
        self.val_progress_bar.close()



import torch
import numpy as np


def to_np(t):
    if t is None:
        return None
    elif t.nelement() == 0:
        return np.array([])
    else:
        return t.cpu().detach().numpy()


def np_unsqueeze(x, dim=0):
    """
    Simulate torch.unsqueeze
    """
    return np.expand_dims(x, axis=dim)




import torch
import functools
import collections.abc
from .core import ExtendedModule, LightningModule


class _Cache:
    """
    Hold training cache
    """
    pass


class MultiOptimizerModule(ExtendedModule):
    """
    Subclass should still have only `hparams` in the arg, and call
    super().__init__(hparams, ...)

    configure_optimizer() method should return a dict instead of a list of optimizers:
    order should be respected
    { 'actor': SGD(), 'critic1': Adam(), 'critic2': Adam() }
    """

    def __new__(cls, *args, **kwargs):
        """
        Class-level patch to avoid multiprocessing pickle error:
            _pickle.PicklingError: Can't pickle <function .. at ..>: it's not
            the same object as __main__.<obj>
        """
        obj = super().__new__(cls)
        cls._patch_configure_optimizer()
        return obj

    def __init__(self, hparams, group_spec):
        """
        group_spec:
        {
            'actor': {
                'interval': 1,
                'training_step_method': self.actor_training_step,
                'optimizer_step_method': <optional>
            },
            'critic1': {
                'interval': 4,
                'training_step_method': self.critic1_training_step,
            },
            'critic2': {
                'interval': 8,
                'training_step_method': self.critic2_training_step,
            }
        }

        NOTE:
          - the optimizer index order will be the same as timescale_spec dict key order
          - training_step_method signature: f(batch, batch_idx) -> output dict
          - optimizer_step_method signature: f(epoch, batch_idx, optimizer, second_order_closure) -> None

        """
        super().__init__(hparams)
        self._group_ordering = []
        assert (
            len(group_spec) >= 2
        ), "must have at least two optimizers to use MultiOptimizerModule"
        for name, conf in group_spec.items():
            assert "interval" in conf and isinstance(conf["interval"], int)
            assert "training_step_method" in conf and callable(
                conf["training_step_method"]
            )
            if "optimizer_step_method" in conf:
                assert callable(conf["optimizer_step_method"])
            else:
                conf["optimizer_step_method"] = None
            self._group_ordering.append(name)
        self.group_spec = group_spec
        self.training_cache = _Cache()
        # dummy loss for skipping training steps
        self._dummy_loss = torch.tensor(0.0, requires_grad=True)

    @property
    def dummy_loss(self):
        return self._dummy_loss.to(self.device)

    def training_step(self, batch, batch_idx, optimizer_idx):
        assert optimizer_idx < len(self._group_ordering), "INTERNAL"
        name = self._group_ordering[optimizer_idx]
        if optimizer_idx == 0:  # first iter over optimizer
            self.training_cache = _Cache()  # reset cache
            self.common_training_step(batch, batch_idx)
        if self.is_training_step_triggered(
            name=name, batch_idx=batch_idx, optimizer_idx=optimizer_idx
        ):
            # self.log_debug('TRIGGER_train_step', name, batch_idx)
            _training_step = self.group_spec[name]["training_step_method"]
            return _training_step(batch, batch_idx)
        else:
            # return a dummy value required by LightningModule
            return {"loss": self._dummy_loss}

    def optimizer_step(
        self,
        epoch: int,
        batch_idx: int,
        optimizer,
        optimizer_idx: int,
        second_order_closure=None,
    ) -> None:
        assert optimizer_idx < len(self._group_ordering), "INTERNAL"
        name = self._group_ordering[optimizer_idx]
        if self.is_training_step_triggered(
            name=name, batch_idx=batch_idx, optimizer_idx=optimizer_idx
        ):
            _optimizer_step = self.group_spec[name]["optimizer_step_method"]
            # self.log_debug('TRIGGER_optimizer_step', name, batch_idx)
            if _optimizer_step is None:
                # default implementation: does not use optimizer_idx in the code
                LightningModule.optimizer_step(
                    self,
                    epoch=epoch,
                    batch_idx=batch_idx,
                    optimizer=optimizer,
                    optimizer_idx=optimizer_idx,
                    second_order_closure=second_order_closure,
                )
            else:
                _optimizer_step(
                    epoch=epoch,
                    batch_idx=batch_idx,
                    optimizer=optimizer,
                    second_order_closure=second_order_closure,
                )

    def is_training_step_triggered(
        self, *, name: str, batch_idx: int, optimizer_idx: int
    ):
        """
        User can override this method to trigger at arbitrary times

        Args:
            name: optimizer group name
        """
        return batch_idx % self.group_spec[name]["interval"] == 0

    def common_training_step(self, batch, batch_idx):
        """
        User can override this method, will be called before optimizer_idx == 0
        Will be called once and only once.
        Write any intermediate saving to self.training_cache
        """
        return None

    @classmethod
    def _patch_configure_optimizer(cls):
        pl_method = cls.configure_optimizers

        @functools.wraps(pl_method)
        def _wrapped(self: MultiOptimizerModule):
            ret = pl_method(self)
            if isinstance(ret, tuple) and len(ret) == 2:
                optimizer_cfg, scheduler_cfg = ret
            else:
                optimizer_cfg, scheduler_cfg = ret, None
            if not isinstance(optimizer_cfg, collections.abc.Mapping):
                raise RuntimeError(
                    f"MultiOptimizerModule: configure_optimizer must return a dictionary of optimizers with the name you specified in `group_spec` at __init__"
                )
            if set(optimizer_cfg.keys()) != set(self.group_spec.keys()):
                raise KeyError(
                    f"Optimizer names {optimizer_cfg.keys()} do not match with `group_spec` names {self.group_spec.keys()}"
                )

            # we return Lightning-compatible list of optimziers in the order of timescale_spec
            optimizer_cfg = [optimizer_cfg[key] for key in self.group_spec]

            if scheduler_cfg is None:
                return optimizer_cfg
            else:
                return optimizer_cfg, scheduler_cfg

        # monkey patch
        cls.configure_optimizers = _wrapped



from . import omlet_logger as _log
from pytorch_lightning.loggers import TensorBoardLogger, WandbLogger
from typing import Dict, Optional


class ExtendedTensorBoardLogger(TensorBoardLogger):
    """
    Tensorboard allows multiple global steps, as long as they are consistent
    """

    def log_metrics(
        self, metrics: Dict[str, float], step: Optional[int] = None
    ) -> None:
        metrics = metrics.copy()
        if "system/epoch" in metrics:
            if "epoch" in metrics:
                del metrics["epoch"]
            if step == metrics["system/epoch"]:
                # only keep `system/global_step` against step (== epoch)
                del metrics["system/epoch"]
            elif step == metrics["system/global_step"]:
                # only keep `system/epoch` against step (== global_step)
                del metrics["system/global_step"]
            else:
                raise ValueError(
                    f"INTERNAL: step {step} is neither "
                    f'global_step {metrics["system/global_step"]} '
                    f'nor epoch {metrics["system/epoch"]}'
                )
        super().log_metrics(metrics, step)
        # _log.debug2(f'Extended TB: {metrics}  step={step}')


class ExtendedWandbLogger(WandbLogger):
    """
    wandb only allows a single `global_step`, so we ignore any other `step`
    """

    def log_metrics(
        self, metrics: Dict[str, float], step: Optional[int] = None
    ) -> None:
        metrics = metrics.copy()
        if "system/global_step" in metrics:
            step = metrics.pop("system/global_step")
            metrics["epoch"] = metrics.pop("system/epoch")
        super().log_metrics(metrics, step)
        # _log.debug2(f'Extended Wandb: {metrics}  step={step}')



import os
import torch
import torch.distributed as _dist
import torch.utils.data
from typing import Union, Optional
from .utils import get_reduce_op, get_backend, _get_group, is_master


def reduce(
    tensor, op="sum", group=None, broadcast=True, out: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """
    Args:
        tensor
        broadcast: True for all-reduce, False for master-only
        out:
            - if the same tensor as input, will mutate input
            - None: create a new tensor and leave the input unchanged
    """
    assert torch.is_tensor(tensor)
    if out is None:
        out = tensor.clone().detach()
    else:
        assert tensor.size() == out.size() and tensor.dtype == out.dtype
    group = _get_group(group)
    opcode = get_reduce_op(op)
    if broadcast:
        _dist.all_reduce(out, op=opcode, group=group)
    else:
        _dist.reduce(out, dst=0, op=opcode, group=group)
    if op.lower() == "mean" and (broadcast or is_master(group)):
        out.div_(1.0 * _dist.get_world_size(group))
    return out


def all_reduce_(tensor, op="sum", group=None):
    return reduce(tensor, op=op, group=group, broadcast=True, out=tensor)


def all_reduce(tensor, op="sum", group=None):
    return reduce(tensor, op=op, group=group, broadcast=True, out=None)


def master_reduce_(tensor, op="sum", group=None):
    return reduce(tensor, op=op, group=group, broadcast=False, out=tensor)


def master_reduce(tensor, op="sum", group=None):
    return reduce(tensor, op=op, group=group, broadcast=False, out=None)


def reduce_scalar(
    scalar, op="sum", *, broadcast=True, storage=None, group=None, device=None
):
    """
    Args:
        scalar: single python float
        broadcast: True to use all_reduce, else master_reduce
        storage: if you call this method repeatedly, consider allocating
            a persistent CUDA tensor in advance to provide temporary storage
            if None, allocates new CUDA memory every time
        device: if None, defaults to torch.cuda.current_device()
    Returns:
        python float
    """
    scalar = float(scalar)
    group = _get_group(group)
    is_nccl = get_backend(group) == "nccl"
    if storage is None:
        s = torch.tensor([scalar], dtype=torch.float32)
        if is_nccl:
            if device is None:
                device = torch.cuda.current_device()
            s = s.cuda(device)
    else:
        assert torch.is_tensor(storage)
        assert storage.numel() > 0, "storage must have at least 1 element"
        if is_nccl:
            assert storage.is_cuda
        storage[0] = scalar
        s = storage
    reduce(s, op=op, group=group, broadcast=broadcast, out=s)
    return float(s[0])


def reduce_scalars(
    scalars, op="sum", *, broadcast=True, storage=None, group=None, device=None
):
    """
    Args:
        scalars: list, tuple or dict of python floats
        broadcast: True to use all_reduce, else master_reduce
        storage: if you call this method repeatedly, consider allocating
            a persistent CUDA tensor in advance to provide temporary storage
            if None, allocates new CUDA memory every time
        device: if None, defaults to torch.cuda.current_device()
    Returns:
        list, tuple or dict of reduced floats
    """
    if not isinstance(scalars, (tuple, list, dict)):
        # singleton falls back to reduce_scalar()
        return reduce_scalar(
            scalars,
            op=op,
            broadcast=broadcast,
            storage=storage,
            group=group,
            device=device,
        )

    group = _get_group(group)
    is_nccl = get_backend(group) == "nccl"
    idx_map = None
    if isinstance(scalars, dict):
        values = []
        idx_map = {}
        for i, (k, v) in enumerate(scalars.items()):
            idx_map[k] = i
            values.append(float(v))
    else:
        values = [float(v) for v in scalars]
    numel = len(values)

    values = torch.tensor(values, dtype=torch.float32)
    if storage is None:
        if is_nccl:
            if device is None:
                device = torch.cuda.current_device()
            values = values.cuda(device)
    else:
        assert torch.is_tensor(storage)
        assert storage.numel() >= numel, "storage not enough"
        if is_nccl:
            assert storage.is_cuda
        storage[:numel] = values
        values = storage
    reduce(values, op=op, group=group, broadcast=broadcast, out=values)

    if isinstance(scalars, dict):
        return {k: float(values[idx_map[k]]) for k in idx_map}
    else:
        return type(scalars)(float(v) for v in values[:numel])



import os
from setuptools import setup, find_packages


def read_file(fname):
    this_dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_dir, fname)) as f:
        return f.read()


pkg_name = 'fovea'


setup(
    name=pkg_name,
    version='0.1.1',
    author='fovea developers',
    # url='http://github.com/',
    description='research project',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    keywords=['Deep Learning',
              'Machine Learning'],
    license='MIT',
    packages=[
        package for package in find_packages() if package.startswith(pkg_name)
    ],
    entry_points={
        'console_scripts': [
            # 'cmd_tool=mylib.subpkg.module:main',
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    python_requires='>=3.7',
    include_package_data=True,
    zip_safe=False
)



from .logging_utils import *
from .config_utils import *
from .print_utils import *
from .torch_utils import *
from .timer import Timer
from .file_utils import *
from .metrics import *
from .misc_utils import *
from .video_utils import *
from .json_yaml_utils import *
from .functional_utils import *
from .np_utils import *
from .convert_utils import *
from .tracer import Tracer

try:
    import tree  # dm_tree
    from .tree_utils import *
except ImportError:
    pass

# FIXME
from ._legacy_utils import *



"""
JSON, YAML, and python config file utilities
"""
import json
import yaml
import os.path as path
from io import StringIO


__all__ = [
    "json_load",
    "json_loads",
    "yaml_load",
    "yaml_loads",
    "json_dump",
    "json_dumps",
    "yaml_dump",
    "yaml_dumps",
    "json_or_yaml_load",
    "json_or_yaml_dump",
]


def json_load(file_path, **kwargs):
    file_path = path.expanduser(file_path)
    with open(file_path, "r") as fp:
        return json.load(fp, **kwargs)


def json_loads(string, **kwargs):
    return json.loads(string, **kwargs)


def json_dump(data, file_path, **kwargs):
    file_path = path.expanduser(file_path)
    with open(file_path, "w") as fp:
        indent = kwargs.pop("indent", 4)
        json.dump(data, fp, indent=indent, **kwargs)


def json_dumps(data, **kwargs):
    "Returns: string"
    return json.dumps(data, **kwargs)


def yaml_load(file_path, *, loader=yaml.safe_load, **kwargs):
    file_path = path.expanduser(file_path)
    with open(file_path, "r") as fp:
        return loader(fp, **kwargs)


def yaml_loads(string, *, loader=yaml.safe_load, **kwargs):
    return loader(string, **kwargs)


def yaml_dump(data, file_path, *, dumper=yaml.safe_dump, **kwargs):
    file_path = path.expanduser(file_path)
    indent = kwargs.pop("indent", 2)
    default_flow_style = kwargs.pop("default_flow_style", False)
    with open(file_path, "w") as fp:
        dumper(
            data,
            stream=fp,
            indent=indent,
            default_flow_style=default_flow_style,
            **kwargs
        )


def yaml_dumps(data, *, dumper=yaml.safe_dump, **kwargs):
    "Returns: string"
    stream = StringIO()
    indent = kwargs.pop("indent", 2)
    default_flow_style = kwargs.pop("default_flow_style", False)
    dumper(data, stream, indent=indent, default_flow_style=default_flow_style, **kwargs)
    return stream.getvalue()


# ==================== auto-recognize extension ====================
def json_or_yaml_load(file_path, **loader_kwargs):
    """
    Args:
        file_path: JSON or YAML loader depends on the file extension

    Raises:
        IOError: if extension is not ".json", ".yml", or ".yaml"
    """
    if file_path.endswith(".json"):
        return json_load(file_path, **loader_kwargs)
    elif file_path.endswith(".yml") or file_path.endswith(".yaml"):
        return yaml_load(file_path, **loader_kwargs)
    else:
        raise IOError(
            'unknown file extension: "{}", loader supports only ".json", ".yml", ".yaml"'.format(
                file_path
            )
        )


def json_or_yaml_dump(data, file_path, **dumper_kwargs):
    """
    Args:
        file_path: JSON or YAML loader depends on the file extension

    Raises:
        IOError: if extension is not ".json", ".yml", or ".yaml"
    """
    if file_path.endswith(".json"):
        return json_dump(data, file_path, **dumper_kwargs)
    elif file_path.endswith(".yml") or file_path.endswith(".yaml"):
        return yaml_dump(data, file_path, **dumper_kwargs)
    else:
        raise IOError(
            'unknown file extension: "{}", dumper supports only ".json", ".yml", ".yaml"'.format(
                file_path
            )
        )



"""
From https://github.com/SsnL/dataset-distillation/blob/master/utils/distributed.py
"""
import torch
import torch.distributed as dist


__all__ = [
    "all_reduce_coalesced",
    "all_gather_coalesced",
    "broadcast_coalesced",
    "barrier",
]


if not dist.is_available():
    # distributed is only supported on Linux, make the following dummy impls for
    # running on macOS & Windows.

    def _coalesce_dummy(*args, **kwargs):
        raise RuntimeError(
            "Trying to run distributed code while torch.distributed is not available. "
            "Make sure that you are on Linux."
        )

    broadcast_coalesced = all_reduce_coalesced = _coalesce_dummy
    all_gather_coalesced = barrier = _coalesce_dummy

else:

    try:
        from torch.distributed import ReduceOp
    except ImportError:
        from torch.distributed import reduce_op

        ReduceOp = reduce_op

    # NB: this might be broken by a future pytorch release
    from torch._utils import (
        _flatten_dense_tensors,
        _unflatten_dense_tensors,
        _take_tensors,
    )

    MB = 1024 * 1024

    def broadcast_coalesced(tensors, src=0, buffer_size=10 * MB):
        r"""
        Broadcast a sequence of tensors to the default group from rank 0.
        Small tensors are first coalesced into a buffer to reduce the number of
        broadcasts.
        tensors (sequence): tensors to broadcast. Each tensor needs to be on the
                            same GPU.
        src (int): src rank. Default: 0.
        buffer_size (int): maximum size of the buffer for coalescing. Default: 10MB.
        """
        for tensors in _take_tensors(tensors, buffer_size):
            flat_tensors = _flatten_dense_tensors(tensors)
            dist.broadcast(flat_tensors, src)
            for old_t, new_t in zip(
                tensors, _unflatten_dense_tensors(flat_tensors, tensors)
            ):
                old_t.data = new_t

    def all_reduce_coalesced(tensors, divisor=1, op=ReduceOp.SUM, buffer_size=256 * MB):
        for tensors in _take_tensors(tensors, buffer_size):
            flat_tensors = _flatten_dense_tensors(tensors)
            dist.all_reduce(flat_tensors, op)
            if divisor != 1:
                flat_tensors.div_(divisor)
            for old_t, new_t in zip(
                tensors, _unflatten_dense_tensors(flat_tensors, tensors)
            ):
                old_t.data = new_t

    # NB: neither nccl nor gloo supports gather, but they are the only options for
    #     gpu distributed, so all gather is all we have

    def all_gather_coalesced(tensors, buffer_size=256 * MB):
        assert (
            dist.get_backend() == dist.dist_backend.NCCL
        )  # gloo gives some weird device error
        world_size = dist.get_world_size()
        rcv_lsts = [[] for _ in range(world_size)]
        for tensors in _take_tensors(tensors, buffer_size):
            flat_tensors = _flatten_dense_tensors(tensors)
            tmp_rcv_lst = [torch.empty_like(flat_tensors) for _ in range(world_size)]
            dist.all_gather(tmp_rcv_lst, flat_tensors)
            for i, rcv_flat_tensors in enumerate(tmp_rcv_lst):
                for rcv_t in _unflatten_dense_tensors(rcv_flat_tensors, tensors):
                    rcv_lsts[i].append(rcv_t)
        return rcv_lsts

    def barrier():
        t = torch.randn((), device="cuda")
        dist.all_reduce(t)
        torch.cuda.synchronize()
